use std::sync::Arc;

use axum::{
    extract::{Json, Path},
    http::StatusCode,
    response::{IntoResponse, Response},
    Extension,
};
use tracing::debug;

use crate::state::ApiState;
use ubisync_lib::{
    api::element::{
        ElementCreateRequest, ElementCreateResponse, ElementGetResponse, ElementSetRequest,
        ElementSetResponse, ElementRemoveResponse,
    },
    types::ElementId,
};

pub(super) async fn get(Path(id): Path<ElementId>, s: Extension<Arc<ApiState>>) -> Response {
    let element = s.get_element(&id);
    match element {
        Ok(el) => (
            StatusCode::OK,
            Json {
                0: ElementGetResponse { element: el },
            },
        )
            .into_response(),
        Err(_) => {
            StatusCode::NOT_FOUND.into_response()
        }
    }
}

pub(super) async fn create(
    s: Extension<Arc<ApiState>>,
    Json(req): Json<ElementCreateRequest>,
) -> Response {
    let element_id = s.create_element(&req.content);
    debug!("{:?}", element_id);
    match element_id {
        Ok(id) => (
            StatusCode::OK,
            Json {
                0: ElementCreateResponse { id },
            },
        )
            .into_response(),
        Err(_) => StatusCode::INTERNAL_SERVER_ERROR.into_response(),
    }
}

pub(super) async fn set(
    Path(id): Path<ElementId>,
    s: Extension<Arc<ApiState>>,
    Json(req): Json<ElementSetRequest>,
) -> Response {
    let res = s.write_element_content(&id, &req.content);
    match res {
        Ok(_) => (
            StatusCode::OK,
            Json {
                0: ElementSetResponse,
            },
        )
            .into_response(),
        Err(_) => StatusCode::INTERNAL_SERVER_ERROR.into_response(),
    }
}

pub(super) async fn remove(Path(id): Path<ElementId>, s: Extension<Arc<ApiState>>) -> Response {
    let res = s.remove_element(&id);
    match res {
        Ok(_) => (StatusCode::OK, Json { 0: ElementRemoveResponse }).into_response(),
        Err(_) => StatusCode::INTERNAL_SERVER_ERROR.into_response(),
    }
}
