use std::collections::BTreeMap;

use anyhow::{bail, Error};
use cozo::{DataValue, DbInstance, JsonData, ScriptMutability};
use serde_json::Value;
use tracing::{debug, error};

use crate::{
    run_query,
    state::{
        Element, ElementContent, ElementId,
    },
};

use ubisync_lib::types::{MessageId, Tag};

pub fn add(
    db: &DbInstance,
    id: &ElementId,
    content: &ElementContent,
    latest_message: Option<MessageId>,
    local_changes: bool,
) -> anyhow::Result<()> {
    let params = vec![
        ("id", DataValue::Str(serde_json::to_string(&id)?.into())),
        (
            "content",
            DataValue::Json(JsonData(serde_json::to_value(content)?)),
        ),
        (
            "latest_message",
            match latest_message {
                Some(m) => DataValue::Str(serde_json::to_string(&m)?.into()),
                None => DataValue::Null,
            },
        ),
        ("local_changes", DataValue::Bool(local_changes)),
    ];

    match run_query!(
        &db,
        ":insert elements {id => content, latest_message, local_changes}",
        params,
        ScriptMutability::Mutable
    ) {
        Ok(_) => Ok(()),
        Err(report) => bail!(report),
    }
}

pub fn set_content(
    db: &DbInstance,
    id: &ElementId,
    content: &ElementContent,
) -> anyhow::Result<()> {
    set_property(
        db,
        id,
        "content",
        DataValue::Json(JsonData(serde_json::to_value(content)?)),
    )
}

pub fn set_latest_message(
    db: &DbInstance,
    id: &ElementId,
    latest_message: Option<MessageId>,
) -> anyhow::Result<()> {
    set_property(
        db,
        id,
        "latest_message",
        match latest_message {
            Some(m) => DataValue::Str(serde_json::to_string(&m)?.into()),
            None => DataValue::Null,
        },
    )
}

pub fn set_local_changes(
    db: &DbInstance,
    id: &ElementId,
    local_changes: bool,
) -> anyhow::Result<()> {
    set_property(db, id, "local_changes", DataValue::Bool(local_changes))
}

pub fn remove(db: &DbInstance, id: &ElementId) -> anyhow::Result<()> {
    match run_query!(
        &db,
        ":delete elements {id}",
        vec![("id", DataValue::Str(serde_json::to_string(&id)?.into()))],
        cozo::ScriptMutability::Mutable
    ) {
        Ok(_) => Ok(()),
        Err(report) => bail!(report),
    }
}

pub fn get(db: &DbInstance, id: &ElementId) -> anyhow::Result<Element> {
    let mut params = BTreeMap::new();
    params.insert(
        "id".to_string(),
        DataValue::Str(serde_json::to_string(&id)?.into()),
    );

    let result = db.run_script("
    ?[content, latest_message, local_changes] := *elements[$id, content, latest_message, local_changes]
    ", params, cozo::ScriptMutability::Immutable);
    match result {
        Ok(val) => {
            if let Some(firstrow) = val.rows.first() {
                debug!("db result: {:?}", &firstrow.as_slice());
                if let [DataValue::Json(JsonData(content)), latest_message, DataValue::Bool(local_changes)] =
                    firstrow.as_slice()
                {
                    return Ok(Element::from((
                        id.to_owned(),
                        serde_json::from_value(content.to_owned())?,
                        match latest_message {
                            DataValue::Str(s) => Some(serde_json::from_str(s)?),
                            _ => None,
                        },
                        local_changes.to_owned(),
                    )));
                }
                return Err(Error::msg("Could not parse db result as Element"));
            } else {
                return Err(Error::msg("No rows returned for element query"));
            }
        }
        Err(report) => bail!(report),
    }
}

pub fn get_by_tag(db: &DbInstance, tag: &Tag) -> anyhow::Result<Vec<ElementId>> {
    let mut params = BTreeMap::new();
    params.insert(
        "tag".to_string(),
        DataValue::Str(serde_json::to_string(tag)?.into()),
    );

    let result = db.run_script(
        "
    ?[element] := *tags[$tag, element]
    ",
        params,
        cozo::ScriptMutability::Immutable,
    );

    match result {
        Ok(named_rows) => {
            let mut element_ids = vec![];
            for row in named_rows.rows {
                if let [DataValue::Json(JsonData(Value::String(element_id)))] = row.as_slice() {
                    match serde_json::from_str(&element_id) {
                        Ok(id) => element_ids.push(id),
                        Err(e) => {
                            error!("Error parsing element id {}: {}", element_id, e);
                            continue;
                        }
                    }
                }
            }
            Ok(element_ids)
        }
        Err(report) => bail!(report),
    }
}

fn set_property(
    db: &DbInstance,
    id: &ElementId,
    key: &str,
    value: DataValue,
) -> anyhow::Result<()> {
    let params = vec![
        ("id", DataValue::Str(serde_json::to_string(id)?.into())),
        (key, value),
    ];

    match run_query!(
        &db,
        format!(":update elements {{id => {key}}}"),
        params,
        ScriptMutability::Mutable
    ) {
        Ok(_) => Ok(()),
        Err(report) => bail!(report),
    }
}
